@@ -1,4 +1,4 @@
-# Copyright (c) 2016 The Regents of the University of Michigan
+# Copyright (c) 2017 The Regents of the University of Michigan
 # All rights reserved.
 # This software is licensed under the BSD 3-Clause License.
 from __future__ import print_function
@@ -9,6 +9,8 @@ import json
 import errno
 import warnings
 import collections
+import shutil
+from itertools import chain
 
 from ..core.search_engine import DocumentSearchEngine
 from ..common import six
@@ -19,6 +21,7 @@ from .indexing import _index_signac_project_workspace
 from .indexing import SignacProjectCrawler
 from .indexing import MasterCrawler
 from .utility import _mkdir_p, is_string
+from .errors import DestinationExistsError
 
 if six.PY2:
     from collections import Mapping
@@ -99,8 +102,7 @@ class JobSearchIndex(object):
         if doc_filter is not None:
             f.update(doc_filter)
         f = json.loads(json.dumps(f))  # Normalize
-        for job_id in self._engine.find(filter=f):
-            yield job_id
+        return self._engine.find(filter=f)
 
 
 class Project(object):
@@ -174,12 +176,27 @@ class Project(object):
                 "Are you sure '{}' is a signac project path?".format(
                     os.path.abspath(self.config.get('project_dir', os.getcwd()))))
 
+    def min_len_unique_id(self):
+        "Determine the minimum length required for an id to be unique."
+        job_ids = list(self.find_job_ids())
+        tmp = set()
+        for i in range(32):
+            tmp.clear()
+            for _id in job_ids:
+                if _id[:i] in tmp:
+                    break
+                else:
+                    tmp.add(_id[:i])
+            else:
+                break
+        return i
+
     def open_job(self, statepoint=None, id=None):
         """Get a job handle associated with a statepoint.
 
         This method returns the job instance associated with
         the given statepoint or job id.
-        Opening a job by statepoint never fails.
+        Opening a job by a valid statepoint never fails.
         Opening a job by id, requires a lookup of the statepoint
         from the job id, which may fail if the job was not
         previously initialized.
@@ -189,8 +206,11 @@ class Project(object):
         :param id: The job id.
         :type id: str
         :return: The job instance.
-        :rtype: :class:`signac.contrib.job.Job`
-        :raises KeyError: If the attempt to open the job by id fails.
+        :rtype: :class:`~.Job`
+        :raises KeyError:
+            If the attempt to open the job by id fails.
+        :raises LookupError: If the attempt to open the job by an
+            abbreviated id returns more than one match.
         """
         if (id is None) == (statepoint is None):
             raise ValueError(
@@ -198,16 +218,23 @@ class Project(object):
         if id is None:
             return self.Job(project=self, statepoint=statepoint)
         else:
+            if len(id) < 32:
+                job_ids = self.find_job_ids()
+                matches = [_id for _id in job_ids if _id.startswith(id)]
+                if len(matches) == 1:
+                    id = matches[0]
+                elif len(matches) > 1:
+                    raise LookupError(id)
             return self.Job(project=self, statepoint=self.get_statepoint(id))
 
     def _job_dirs(self):
         wd = self.workspace()
-        m = re.compile('[a-z0-9]{32}')
+        m = re.compile('[a-f0-9]{32}')
         try:
             for d in os.listdir(wd):
                 if m.match(d):
                     yield d
-        except IOError as error:
+        except OSError as error:
             if error.errno != errno.ENOENT:
                 raise
 
@@ -215,6 +242,18 @@ class Project(object):
         "Return the number of initialized jobs."
         return len(list(self._job_dirs()))
 
+    __len__ = num_jobs
+
+    def __contains__(self, job):
+        """Determine whether job is in the project's data space.
+
+        :param job: The job to test for initialization.
+        :type job: :py:class:`~.Job`
+        :returns: True when the job is initialized for this project.
+        :rtype: bool
+        """
+        return job.get_id() in self.find_job_ids()
+
     def build_job_search_index(self, index, include=None, hash_=None):
         """Build a job search index.
 
@@ -288,9 +327,7 @@ class Project(object):
             by the index.
         """
         if filter is None and doc_filter is None and index is None:
-            for job_id in self._job_dirs():
-                yield job_id
-            return
+            return list(self._job_dirs())
         if index is None:
             index = self.index(include_job_document=doc_filter is not None)
         if doc_filter is None:
@@ -298,8 +335,7 @@ class Project(object):
         else:
             include = None
         search_index = self.build_job_search_index(index, include)
-        for job_id in search_index.find_job_ids(filter=filter, doc_filter=doc_filter):
-            yield job_id
+        return search_index.find_job_ids(filter=filter, doc_filter=doc_filter)
 
     def find_jobs(self, filter=None, doc_filter=None, index=None):
         """Find all jobs in the project's workspace.
@@ -322,8 +358,10 @@ class Project(object):
         :raises RuntimeError: If the filters are not supported
             by the index.
         """
-        for job_id in self.find_job_ids(filter, doc_filter, index):
-            yield self.open_job(id=job_id)
+        return _JobsIterator(self, self.find_job_ids(filter, doc_filter, index))
+
+    def __iter__(self):
+        return self.find_jobs()
 
     def find_statepoints(self, filter=None, doc_filter=None, index=None, skip_errors=False):
         """Find all statepoints in the project's workspace.
@@ -345,38 +383,6 @@ class Project(object):
         for job in jobs:
             yield job.statepoint()
 
-    def find_variable_parameters(self, statepoints=None):
-        """Find all parameters which vary over the data space.
-
-        .. warning::
-
-            This method is deprecated.
-            Please see :meth:`~.build_job_statepoint_index` for an
-            alternative method.
-
-        This method attempts to detect all parameters, which vary
-        over the parameter space.
-        The parameter sets are ordered decreasingly
-        by data sub space size.
-
-        .. warning::
-
-            This method does not detect linear dependencies
-            within the state points. Linear dependencies should
-            generally be avoided.
-
-        :param statepoints: The statepoints to consider.
-            Defaults to all state points within the data space.
-        :type statepoints: Iterable of parameter mappings.
-        :return: A hierarchical list of variable parameters.
-        :rtype: list"""
-        warnings.warn(
-            "The find_variable_parameters() method is deprecated, please use "
-            "build_job_statepoint_index() instead.", DeprecationWarning)
-        if statepoints is None:
-            statepoints = self.find_statepoints()
-        return list(_find_unique_keys(statepoints))
-
     def read_statepoints(self, fn=None):
         """Read all statepoints from a file.
 
@@ -484,9 +490,8 @@ class Project(object):
         assert str(self.open_job(statepoint)) == jobid
         return statepoint
 
-    def create_linked_view(self, job_ids=None, prefix=None,
-                           force=False, index=None):
-        """Create a persistent linked view of the selected data space..
+    def create_linked_view(self, prefix=None, job_ids=None, index=None):
+        """Create or update a persistent linked view of the selected data space.
 
         This method determines unique paths for each job based on the job's
         statepoint and creates symbolic links to the associated workspace
@@ -513,97 +518,37 @@ class Project(object):
 
             To maximize the compactness of each view path, *b* which does not
             vary over the selected data space, is ignored.
+
+        :param prefix:
+            The path where the linked view will be created or updated.
+        :type prefix:
+            str
+        :param job_ids:
+            If None (the default), create the view for the complete data space,
+            otherwise only for the sub space constituted by the provided job ids.
+        :param index:
+            A document index.
         """
         if prefix is None:
             prefix = 'view'
         if index is None:
             index = self.index(include_job_document=False)
-        if not force and os.listdir(prefix):
-            raise RuntimeError(
-                "Failed to create persistent view in '{}', the directory "
-                "is not empty! Use `force=True` to ignore this and create "
-                "the view anyways.".format(prefix))
-
         if job_ids is not None:
             if not isinstance(job_ids, set):
                 job_ids = set(job_ids)
-            index = (doc for doc in index if doc['signac_id'] in job_ids)
+            index = [doc for doc in index if doc['_id'] in job_ids]
+            if not job_ids.issubset({doc['_id'] for doc in index}):
+                raise ValueError("Insufficient index for selected data space.")
 
         jsi = self.build_job_statepoint_index(exclude_const=True, index=index)
-        no_link = True
+        links = dict()
         for path, job_id in _make_paths(jsi):
-            if job_ids is not None and job_id not in job_ids:
-                continue
-            src = os.path.join(self.open_job(id=job_id).workspace())
-            dst = os.path.join(prefix, path)
-            logger.info(
-                "Creating link {src} -> {dst}".format(src=src, dst=dst))
-            _make_link(src, dst)
-            no_link = False
-        if no_link:
-            raise RuntimeError(
-                "The # of jobs selected for the creation of views must "
-                "be greater or equal than 2!")
-
-    def create_view(self, filter=None, prefix='view'):
-        """Create a view of the workspace.
-
-        .. warning::
-
-            This method is deprecated.
-            Please use :meth:`~.create_linked_view` instead.
-
-        This method gathers all varying statepoint parameters
-        and creates symbolic links to the workspace directories.
-        This is useful for browsing through the workspace in a
-        human-readable manner.
-
-        Let's assume the parameter space is
-
-            * a=0, b=0
-            * a=1, b=0
-            * a=2, b=0
-            * ...,
-
-        where *b* does not vary over all statepoints.
-
-        Calling this method will generate the following *symbolic links* within
-        the speciefied  view directory:
-
-        .. code-block:: bash
-
-            view/a/0 -> /path/to/workspace/7f9fb369851609ce9cb91404549393f3
-            view/a/1 -> /path/to/workspace/017d53deb17a290d8b0d2ae02fa8bd9d
-            ...
-
-        .. note::
-
-            As *b* does not vary over the whole parameter space it is not part
-            of the view url.
-            This maximizes the compactness of each view url.
-
-        :param filter:  If not None,
-            create view only for jobs matching filter.
-        :type filter: mapping
-        :param prefix: Specifies where to create the links."""
-        warnings.warn(
-            "The create_view() method is deprecated, please use "
-            "create_linked_view() instead.", DeprecationWarning)
-        statepoints = list(self.find_statepoints(filter=filter))
-        if not len(statepoints):
-            if filter is None:
-                logger.warning("No state points found!")
-            else:
-                logger.warning("No state points matched the filter.")
-        key_set = list(_find_unique_keys(statepoints))
-        if filter is not None:
-            key_set[:0] = [[key] for key in filter.keys()]
-        for statepoint, url in _make_urls(statepoints, key_set):
-            src = self.open_job(statepoint).workspace()
-            dst = os.path.join(prefix, url)
-            logger.info(
-                "Creating link {src} -> {dst}".format(src=src, dst=dst))
-            _make_link(src, dst)
+            links[path] = self.open_job(id=job_id).workspace()
+        if not links:   # data space contains less than two elements
+            for job in self.find_jobs():
+                links['./job'] = job.workspace()
+            assert len(links) < 2
+        _update_view(prefix, links)
 
     def find_job_documents(self, filter=None):
         """Find all job documents in the project's workspace.
@@ -633,33 +578,31 @@ class Project(object):
             yield doc
 
     def reset_statepoint(self, job, new_statepoint):
-        """Reset the statepoint of job.
+        """Reset the state point of job.
 
         .. danger::
 
-            Use this function with caution! Resetting a job's statepoint,
+            Use this function with caution! Resetting a job's state point,
             may sometimes be necessary, but can possibly lead to incoherent
             data spaces.
-            If you only want to *extend* your statepoint, consider to
-            use :meth:`~.update_statepoint` instead.
 
         :param job: The job, that should be reset to a new state point.
         :type job: :class:`~.contrib.job.Job`
-        :param new_statepoint: The job's new unique set of parameters.
+        :param new_statepoint: The job's new state point.
         :type new_statepoint: mapping
-        :raises RuntimeError: If a job associated with the new unique set
-            of parameters already exists in the workspace."""
-        dst = self.open_job(new_statepoint)
-        _move_job(job, dst)
-        logger.info(
-            "Reset statepoint of job {}, moved to {}.".format(job, dst))
+        :raises RuntimeError:
+            If a job associated with the new state point is already initialized.
+        :raises OSError:
+            If the move failed due to an unknown system related error.
+        """
+        job.reset_statepoint(new_statepoint=new_statepoint)
 
     def update_statepoint(self, job, update, overwrite=False):
-        """Update the statepoint of job.
+        """Update the statepoint of this job.
 
         .. warning::
 
-            While appending to a job's statepoint is generally safe,
+            While appending to a job's state point is generally safe,
             modifying existing parameters may lead to data
             inconsistency. Use the overwrite argument with caution!
 
@@ -667,20 +610,41 @@ class Project(object):
         :type job: :class:`~.contrib.job.Job`
         :param update: A mapping used for the statepoint update.
         :type update: mapping
-        :param overwrite: Set to true, to ignore whether this
-            update overwrites parameters, which are currently
-            part of the job's statepoint. Use with caution!
-        :raises KeyError: If the update contains keys, which are
-            already part of the job's statepoint and overwrite is False.
-        :raises RuntimeError: If a job associated with the new unique set
-            of parameters already exists in the workspace."""
-        statepoint = dict(job.statepoint())
-        if not overwrite:
-            for key in update:
-                if key in statepoint:
-                    raise KeyError(key)
-        statepoint.update(update)
-        _move_job(job, self.open_job(statepoint))
+        :param overwrite:
+            Set to true, to ignore whether this update overwrites parameters,
+            which are currently part of the job's state point. Use with caution!
+        :raises KeyError:
+            If the update contains keys, which are already part of the job's
+            state point and overwrite is False.
+        :raises RuntimeError:
+            If a job associated with the new state point is already initialized.
+        :raises OSError:
+            If the move failed due to an unknown system related error.
+        """
+        job.update_statepoint(update=update, overwrite=overwrite)
+
+    def clone(self, job):
+        """Clone job into this project.
+
+        Create an identical copy of job within this project.
+
+        :param job: The job to copy into this project.
+        :type job: :py:class:`~.Job`
+        :returns: The job instance corresponding to the copied job.
+        :rtype: :py:class:`~.Job`
+        :raises DestinationExistsError:
+            In case that a job with the same id is already
+            initialized within this project.
+        """
+        dst = self.open_job(job.statepoint())
+        try:
+            shutil.copytree(job.workspace(), dst.workspace())
+        except OSError as error:
+            if error.errno == errno.EEXIST:
+                raise DestinationExistsError(dst)
+            else:
+                raise
+        return dst
 
     def repair(self):
         "Attempt to repair the workspace after it got corrupted."
@@ -889,21 +853,118 @@ class Project(object):
         return cls(config=config)
 
 
-def _move_job(src, dst):
-    logger.debug("Attempting to move job {} to {}".format(src, dst))
-    fn_src_manifest = os.path.join(src.workspace(), src.FN_MANIFEST)
-    fn_src_manifest_backup = fn_src_manifest + '~'
-    os.rename(fn_src_manifest, fn_src_manifest_backup)
-    try:
-        os.rename(src.workspace(), dst.workspace())
-    except OSError:  # rollback
-        os.rename(fn_src_manifest_backup, fn_src_manifest)
-        raise RuntimeError(
-            "Failed to move {} to {}, destination already exists.".format(
-                src, dst))
+def _find_all_links(root, leaf='job'):
+    for dirpath, dirnames, filenames in os.walk(root):
+        for dirname in dirnames:
+            if dirname == leaf:
+                yield os.path.relpath(dirpath, root)
+                break
+        for filename in filenames:
+            if filename == leaf:
+                yield os.path.relpath(dirpath, root)
+                break
+
+
+class _Node(object):
+
+    def __init__(self, name=None, value=None):
+        self.name = name
+        self.value = value
+        self.children = dict()
+
+    def get_child(self, name):
+        return self.children.setdefault(name, type(self)(name))
+
+    def __str__(self):
+        return "_Node({}, {})".format(self.name, self.value)
+
+    __repr__ = __str__
+
+
+def _build_tree(paths):
+    root = _Node()
+    for path in paths:
+        node = root
+        for p in path.split(os.sep):
+            node = node.get_child(p)
+    return root
+
+
+def _get_branches(root, branch=None):
+    if branch is None:
+        branch = list()
     else:
-        dst.init()
-        logger.info("Moved job {} to {}.".format(src, dst))
+        branch = list(branch) + [root]
+    if root.children:
+        for child in root.children.values():
+            for b in _get_branches(child, branch):
+                yield b
+    else:
+        yield branch
+
+
+def _color_path(root, path):
+    root.value = True
+    for name in path:
+        root = root.get_child(name)
+        root.value = True
+
+
+def _find_dead_branches(root, branch=None):
+    if branch is None:
+        branch = list()
+    else:
+        branch = list(branch) + [root]
+    if root.children:
+        for child in root.children.values():
+            for b in _find_dead_branches(child, branch):
+                yield b
+    if not root.value:
+        yield branch
+
+
+def _analyze_view(prefix, links, leaf='job'):
+    logger.info("Analyzing view prefix '{}'...".format(prefix))
+    existing_paths = {os.path.join(p, leaf) for p in _find_all_links(prefix, leaf)}
+    existing_tree = _build_tree(existing_paths)
+    for path in links:
+        _color_path(existing_tree, path.split(os.sep))
+    obsolete = []
+    dead_branches = _find_dead_branches(existing_tree)
+    for branch in reversed(sorted(dead_branches, key=lambda b: len(b))):
+        if branch:
+            obsolete.append(os.path.join(* (n.name for n in branch)))
+    if '.' in obsolete:
+        obsolete.remove('.')
+    keep_or_update = existing_paths.intersection(links.keys())
+    new = set(links.keys()).difference(keep_or_update)
+    to_update = [p for p in keep_or_update if os.path.realpath(os.path.join(prefix, p)) != links[p]]
+    return obsolete, to_update, new
+
+
+def _update_view(prefix, links, leaf='job'):
+    obsolete, to_update, new = _analyze_view(prefix, links)
+    num_ops = len(obsolete) + 2 * len(to_update) + len(new)
+    if num_ops:
+        logger.info("Generating current view in '{}' ({} operations)...".format(prefix, num_ops))
+    else:
+        logger.info("View in '{}' is up to date.".format(prefix))
+        return
+    logger.debug("Removing {} obsolete links.".format(len(obsolete)))
+    for path in obsolete:
+        p = os.path.join(prefix, path)
+        try:
+            os.unlink(p)
+        except OSError:
+            os.rmdir(p)
+    logger.debug("Creating {} new and updating {} existing links.".format(
+        len(new), len(to_update)))
+    for path in to_update:
+        os.unlink(os.path.join(prefix, path))
+    for path in chain(new, to_update):
+        dst = os.path.join(prefix, path)
+        src = os.path.relpath(links[path], os.path.split(dst)[0])
+        _make_link(src, dst)
 
 
 def _make_link(src, dst):
@@ -941,72 +1002,17 @@ def _make_urls(statepoints, key_set):
             yield statepoint, os.path.join(*url)
 
 
-def _make_paths(sp_index):
+def _make_paths(sp_index, leaf='job'):
     tmp = collections.defaultdict(list)
     for key, jids in sp_index:
         for jid in jids:
             tmp[jid].append(key)
     for jid, sps in tmp.items():
         p = ('_'.join(str(x) for x in json.loads(sp)) for sp in sorted(sps))
-        path = os.path.join(* list(p) + ['job'])
+        path = os.path.join(* list(p) + [leaf])
         yield path, jid
 
 
-def _find_unique_keys(statepoints):
-    key_set = _aggregate_statepoints(statepoints)
-    if six.PY2:
-        def flatten(l):
-            for el in l:
-                if isinstance(el, collections.Iterable) and not \
-                        (isinstance(el, str) or isinstance(el, unicode)):  # noqa
-                    for sub in flatten(el):
-                        yield sub
-                else:
-                    yield el
-    else:
-        def flatten(l):
-            for el in l:
-                if isinstance(el, collections.Iterable) and \
-                        not (isinstance(el, str)):
-                    for sub in flatten(el):
-                        yield sub
-                else:
-                    yield el
-    key_set = (list(flatten(k)) for k in key_set)
-    for key in sorted(key_set, key=len):
-        yield key
-
-
-def _aggregate_statepoints(statepoints, prefix=None):
-    result = list()
-    statepoint_set = collections.defaultdict(set)
-    # Gather all keys.
-    ignore = set()
-    for statepoint in statepoints:
-        for key, value in statepoint.items():
-            if key in ignore:
-                continue
-            try:
-                statepoint_set[key].add(value)
-            except TypeError:
-                if isinstance(value, Mapping):
-                    result.extend(_aggregate_statepoints(
-                        [sp[key] for sp in statepoints if key in sp],
-                        prefix=(key) if prefix is None else (prefix, key)))
-                    ignore.add(key)
-                else:
-                    statepoint_set[key].add(calc_id(value))
-    # Heal heterogenous parameter space.
-    for statepoint in statepoints:
-        for key in statepoint_set.keys():
-            if key not in statepoint:
-                statepoint_set[key].add(None)
-    unique_keys = list(k for k, v in sorted(
-        statepoint_set.items(), key=lambda i: len(i[1])) if len(v) > 1)
-    result.extend((k,) if prefix is None else (prefix, k) for k in unique_keys)
-    return result
-
-
 def _skip_errors(iterable, log=print):
     while True:
         try:
@@ -1017,6 +1023,26 @@ def _skip_errors(iterable, log=print):
             log(error)
 
 
+class _JobsIterator(object):
+
+    def __init__(self, project, ids):
+        self._project = project
+        self._ids = ids
+        self._ids_iterator = iter(ids)
+
+    def __len__(self):
+        return len(self._ids)
+
+    def __iter__(self):
+        return self
+
+    def __next__(self):
+        return self._project.open_job(id=next(self._ids_iterator))
+
+    next = __next__  # python 2.7 compatibility
+
+
+
 def init_project(name, root=None, workspace=None, make_dir=True):
     """Initialize a project with the given name.
 
